from os.path import commonpath
from pathlib import Path
from tarfile import open as tarfile_open
from typing import List, Optional, Union


def compress_files(
    filenames: List[Union[str, Path]],
    compressed_file: Union[str, Path],
    arcname: Optional[Union[str, Path]] = None,
) -> bool:
    """
    Given a list of files or directories, compress them into a tarball at the
    compressed_file location.

    Parameters
    ----------
    filenames : List[Union[str, Path]]
        List of files or directories
    compressed_file : Union[str, Path]
        Destination compressed file.
    arcname : Union[str, Path]
        Base directory for the compressed file. Default: Maximum common path for all
        the passed in filenames.

    Returns
    -------
    bool
        The status of the compress operation. Checks if the output file was created.

    Raises
    ------
    NameError
        Raised when the expected name of the compressed file doesn't end in tar or tar.gz.
    """
    # If the file to be written is provided as an str object, convert it to a Path object.
    if isinstance(compressed_file, str):
        compressed_file = Path(compressed_file)

    # Ensure the file to be written ends with tar or tar.gz.
    if compressed_file.name.endswith(".tar.gz") or compressed_file.name.endswith(
        ".tar"
    ):
        tar_file = tarfile_open(compressed_file, "w:gz")
    else:
        raise NameError(
            f"Name of compressed file ({compressed_file}) does not end in tar or tar.gz."
        )

    base_path = Path(commonpath(filenames))

    for filename in filenames:
        if isinstance(filename, str):
            filename = Path(filename)

        if arcname is None:
            arcname = filename.relative_to(base_path)
            # If common path is "", the arcname becomes a directory `.` in which all
            # files are then stored. This line avoids that scenario.
            arcname = "" if arcname.as_posix() == "." else arcname
        tar_file.add(filename, arcname=arcname)
    tar_file.close()
    return compressed_file.exists()



from logging import getLogger
from pathlib import Path
from shutil import copy2
from tempfile import TemporaryDirectory
from typing import Callable, List, Optional, Tuple, Union

from s3fs import S3FileSystem
from torch import Tensor, cat, cuda, device
from torch.autograd.grad_mode import no_grad
from torch.nn.functional import pad
from tqdm.auto import trange
from transformers import (
    AutoModel,
    AutoTokenizer,
    BatchEncoding,
    PreTrainedModel,
    PreTrainedTokenizer,
    PreTrainedTokenizerFast,
)
from transformers.file_utils import ModelOutput

from transformer_embeddings.helpers import compress_files


logger = getLogger(__name__)

MODEL_TARBALL = "model.tar.gz"

DEVICE_CUDA = device("cuda")
DEVICE_CPU = device("cpu")
DEVICE = DEVICE_CUDA if cuda.is_available() else DEVICE_CPU

TransformerInputOutput = Union[BatchEncoding, ModelOutput]


class TransformerEmbeddingsOutput(ModelOutput):
    output: Optional[ModelOutput] = None
    input: Optional[BatchEncoding] = None
    pooled: Optional[Tensor] = None


class TransformerEmbeddings:
    """Thin wrapper on top of the HuggingFace's transformers library to simplify
    generating embeddings."""

    def __init__(
        self,
        model_name: Optional[Union[str, Path]] = None,
        tokenizer: Union[PreTrainedTokenizer, PreTrainedTokenizerFast] = None,
        model: PreTrainedModel = None,
        batch_size: int = 16,
        tensors_to_cpu: bool = True,
        return_output: bool = True,
        return_input: bool = False,
        pooling_fn: Callable[[ModelOutput, BatchEncoding], Tensor] = None,
    ):
        """
        Create a TransformerEmbeddings object.

        Parameters
        ----------
        model_name : Union[str, Path], optional
            Name of the model. Anything supported by HF's `from_pretrained()` method
            (HF model hub model name, local path). Default: None.
        tokenizer : Union[PreTrainedTokenizer, PreTrainedTokenizerFast], optional
            Tokenizer object. Default: None.
        model : PreTrainedModel, optional
            Model object. Default: None.
        batch_size : int, optional
            Batch size for the foward pass. Default: 16.
        tensors_to_cpu : bool, optional
            Move the output back to CPU if this is operating on GPU? Default: True.
        return_output : bool, optional
            Should all the outputs from the model's forward pass be returned? Default: True.
        return_input : bool, optional
            Should the tokenized inputs be returned? Default: False.
        pooling_fn : Callable[[ModelOutput, BatchEncoding], Tensor], optional
            Function to apply to pool the output produced into a tensor. If provided,
            self.return_pooled is set to True and returned.
        """
        self.load_model(model_name=model_name, tokenizer=tokenizer, model=model)

        self.batch_size = batch_size
        self.tensors_to_cpu = tensors_to_cpu
        self.pooling_fn = pooling_fn

        self.return_output = return_output
        self.return_input = return_input
        self.return_pooled = self.pooling_fn is not None

        logger.info("TransformerEmbedddings model initialized.")

    def load_model(
        self,
        model_name: Optional[str],
        tokenizer: PreTrainedTokenizer,
        model: PreTrainedModel,
    ) -> None:
        """
        Load the model.

        Raises
        ------
        ValueError
            If either model or tokenizer are provided without the other, and model_name is also not provided.
        """

        if model_name and tokenizer is None:
            logger.info(f"Loading tokenizer from {model_name}")
            tokenizer = AutoTokenizer.from_pretrained(model_name)

        if model_name and model is None:
            logger.info(f"Loading model from {model_name}")
            model = AutoModel.from_pretrained(model_name)

        if tokenizer is None:
            raise ValueError("Tokenizer was not passed or created.")
        else:
            self.tokenizer = tokenizer

        if model is None:
            raise ValueError("Model was not passed or created.")
        else:
            self.model = model.to(DEVICE).eval()
        logger.info(f"Model and tokenizer loaded, on device {DEVICE}, set to eval().")
        if (
            model.config
            and model.config.max_position_embeddings
            and tokenizer.model_max_length
            and model.config.max_position_embeddings != tokenizer.model_max_length
        ):
            logger.warning(
                f"Model's maximum position embeddings ({model.config.max_position_embeddings}) do not match tokenizer's maximum length ({tokenizer.model_max_length})."
            )

    def tokenize(self, input_strings: List[str]) -> BatchEncoding:
        """
        Tokenize the input.

        Parameters
        ----------
        input_strings : List[str]
            Input strings for the batch.

        Returns
        -------
        BatchEncoding
            Tokenizer output.
        """
        return self.tokenizer(
            input_strings,
            return_tensors="pt",
            padding=True,
            truncation=True,
        )

    def _pad_tensors(self, x: Tensor, y: Tensor) -> Tuple[Tensor, Tensor]:
        """Pad tensors to equal length."""
        max_length = max(x.shape[1], y.shape[1])
        # padding passed to torch.nn.functional.pad are backwards from the last axis
        # and is specified twice for each dimension (front and back).
        # input tensors are 2D, output tensors are 3D.
        # input and output tensors both have to be padded at dimension 1.
        # padding for input tensors becomes: (0, max_length - x.shape[1])
        # padding for output tensors becomes: (0, 0, 0, max_length - x.shape[1])
        # See: https://pytorch.org/docs/master/generated/torch.nn.functional.pad.html
        # We don't pad along the dimension we stack (dim=0; batch dimension), thus
        # the -2. For the final dimension, we only pad at the back, hence the -1.
        prefix_length = 2 * len(x.shape) - 2 - 1
        pad_prefix = (0,) * prefix_length
        x = pad(x, pad=pad_prefix + (max_length - x.shape[1],))
        y = pad(y, pad=pad_prefix + (max_length - y.shape[1],))
        return x, y

    def _change_device(self, data: TransformerInputOutput) -> TransformerInputOutput:
        """Move tensors to CPU if specified."""
        if self.tensors_to_cpu:
            for key, value in data.items():
                data[key] = value.to(DEVICE_CPU)
        return data

    def _stack_batch(
        self,
        data: Optional[TransformerInputOutput],
        batch_data: TransformerInputOutput,
    ) -> TransformerInputOutput:
        """Stack batch data with data from previous batches."""
        # Stack.
        if data is None:
            data = batch_data
        else:
            for key, value in batch_data.items():
                previous = data[key]
                if previous.shape != value.shape:
                    previous, value = self._pad_tensors(previous, value)
                data[key] = cat((previous, value), dim=0)
        return data

    def _stack_pooled(self, pooled: Optional[Tensor], batch_pooled: Tensor) -> Tensor:
        # Pooling always occurs on tensors on the same device, so we do not need to
        # move them. The pooled tensor would already be on CPU if tensors_to_cpu is
        # True, else they'd be on GPU. Either way, they'd be on the device we want it on.
        return batch_pooled if pooled is None else cat((pooled, batch_pooled), dim=0)

    def encode(self, input_strings: List[str]) -> TransformerEmbeddingsOutput:
        """
        Generate embeddings for the given input.

        Parameters
        ----------
        input_strings : List[str]
            String input for which embeddings should be generated.

        Returns
        -------
        TransformerEmbeddingsOutput
            Model output, inputs and / or pooled output.
        """
        logger.info(f"Generating embeddings for {len(input_strings)} input strings.")
        output = None
        input = None
        pooled = None
        for i in trange(0, len(input_strings), self.batch_size):
            batch_tokenized_input = self.tokenize(
                input_strings[i : i + self.batch_size]
            )

            with no_grad():
                batch_outputs = self.model(**batch_tokenized_input.to(DEVICE))

            # Move all tensors to CPU.
            batch_tokenized_input = self._change_device(batch_tokenized_input)
            batch_outputs = self._change_device(batch_outputs)

            # Stack tensors from batch to output for all inputs.
            if self.return_output:
                output = self._stack_batch(output, batch_outputs)
            if self.return_input:
                input = self._stack_batch(input, batch_tokenized_input)
            if self.return_pooled:
                batch_pooled = self.pooling_fn(batch_outputs, batch_tokenized_input)
                pooled = self._stack_pooled(pooled, batch_pooled)
        return TransformerEmbeddingsOutput(output=output, input=input, pooled=pooled)

    def export(
        self,
        output_dir: Optional[Union[str, Path]] = None,
        additional_files: Optional[List[Union[str, Path]]] = None,
        s3_path: Optional[str] = None,
    ) -> Path:
        """
        Export the model and tokenizer to a directory and compress it into a tarball. If
        an S3 path is provided, also upload the tarball to S3.

        Parameters
        ----------
        output_dir : Optional[Union[str, Path]], optional
            Output directory. Default: A temporary directory is created that is cleaned up when the function returns.
        additional_files : Optional[List[Path]], optional
            Additional files to include in the exported tarball. Default: None.
        s3_path : Optional[str], optional
            S3 path at which to upload the file. Default: None, which means that the file is not uploaded to S3.

        Returns
        -------
        Path
            Tarball path. If it was a temporary directory, it will be empty.
        """
        # Set directory.
        temporary_directory = TemporaryDirectory()
        if output_dir is None:
            output_dir = temporary_directory.name
        output_dir = Path(output_dir) if isinstance(output_dir, str) else output_dir
        if not output_dir.exists() or output_dir.is_dir():
            output_dir.mkdir(parents=True, exist_ok=True)

        # Copy additional files into output directory before compressing.
        for file in additional_files or []:
            copy2(file, output_dir)

        # Save model, tokenizer.
        self.model.save_pretrained(output_dir)
        self.tokenizer.save_pretrained(output_dir)

        # Create a tarball.
        logger.debug(f"Folder being added to the tarball is {output_dir}.")
        compressed_file = output_dir.joinpath(MODEL_TARBALL)
        compress_files([output_dir], compressed_file)
        logger.info(f"Tarball {compressed_file} created.")

        if s3_path:
            s3_fs = S3FileSystem()
            logger.info(f"Tarball {compressed_file} being uploaded to S3 at {s3_path}.")
            s3_fs.open(s3_path, "wb").write(compressed_file.read_bytes())

        return compressed_file



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['transformer_embeddings']

package_data = \
{'': ['*']}

install_requires = \
['s3fs>=2022.10.0,<2023.0.0',
 'torch>=1.9.1,<2.0.0',
 'transformers>=4.23.1,<5.0.0']

setup_kwargs = {
    'name': 'transformer-embeddings',
    'version': '3.0.5',
    'description': 'Transformer Embeddings',
    'long_description': '# Transformer Embeddings\n\n[![PyPI](https://img.shields.io/pypi/v/transformer-embeddings.svg)][pypi_]\n[![Status](https://img.shields.io/pypi/status/transformer-embeddings.svg)][status]\n[![Python Version](https://img.shields.io/pypi/pyversions/transformer-embeddings)][python version]\n[![License](https://img.shields.io/pypi/l/transformer-embeddings)][license]\n\n[![Tests](https://github.com/ginger-io/transformer-embeddings/workflows/Tests/badge.svg?branch=main)][tests]\n\n[![pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)][pre-commit]\n[![Black](https://img.shields.io/badge/code%20style-black-000000.svg)][black]\n\n[pypi_]: https://pypi.org/project/transformer-embeddings/\n[status]: https://pypi.org/project/transformer-embeddings/\n[python version]: https://pypi.org/project/transformer-embeddings\n[read the docs]: https://transformer-embeddings.readthedocs.io/\n[tests]: https://github.com/ginger-io/transformer-embeddings/actions?workflow=Tests\n[codecov]: https://app.codecov.io/gh/ginger-io/transformer-embeddings\n[pre-commit]: https://github.com/pre-commit/pre-commit\n[black]: https://github.com/psf/black\n\nThis library simplifies and streamlines the usage of encoder transformer models supported by [HuggingFace\'s `transformers` library](https://github.com/huggingface/transformers/) ([model hub](https://huggingface.co/models) or local) to generate embeddings for string inputs, similar to the way `sentence-transformers` does.\n\n## Why use this over HuggingFace\'s `transformers` or `sentence-transformers`?\n\nUnder the hood, we take care of:\n\n1. Can be used with any model on the HF model hub, with sensible defaults for inference.\n2. Setting the PyTorch model to `eval` mode.\n3. Using `no_grad()` when doing the forward pass.\n4. Batching, and returning back output in the format produced by HF transformers.\n5. Padding / truncating to model defaults.\n6. Moving to and from GPUs if available.\n\n## Installation\n\nYou can install _Transformer Embeddings_ via [pip] from [PyPI]:\n\n```console\n$ pip install transformer-embeddings\n```\n\n## Usage\n\n```python\nfrom transformer_embeddings import TransformerEmbeddings\n\ntransformer = TransformerEmbeddings("model_name")\n```\n\nIf you have a previously instantiated `model` and / or `tokenizer`, you can pass that in.\n\n```python\ntransformer = TransformerEmbeddings(model=model, tokenizer=tokenizer)\n```\n\n```python\ntransformer = TransformerEmbeddings(model_name="model_name", model=model)\n```\n\nor\n\n```python\ntransformer = TransformerEmbeddings(model_name="model_name", tokenizer=tokenizer)\n```\n\n**Note:** The `model_name` should be included if only 1 of model or tokenizer are passed in.\n\n### Embeddings\n\nTo get output embeddings:\n\n```python\nembeddings = transformer.encode(["Lorem ipsum dolor sit amet",\n                                 "consectetur adipiscing elit",\n                                 "sed do eiusmod tempor incididunt",\n                                 "ut labore et dolore magna aliqua."])\nembeddings.output\n```\n\n### Pooled Output\n\nTo get pooled outputs:\n\n```python\nfrom transformer_embeddings import TransformerEmbeddings, mean_pooling\n\ntransformer = TransformerEmbeddings("model_name", return_output=False, pooling_fn=mean_pooling)\n\nembeddings = transformer.encode(["Lorem ipsum dolor sit amet",\n                                "consectetur adipiscing elit",\n                                "sed do eiusmod tempor incididunt",\n                                "ut labore et dolore magna aliqua."])\n\nembeddings.pooled\n```\n\n### Exporting the Model\n\nOnce you are done testing and training the model, it can be exported into a single tarball that can be uploaded to S3 using:\n\n```python\nfrom transformer_embeddings import TransformerEmbeddings\n\ntransformer = TransformerEmbeddings("model_name")\ntransformer.export(\n    additional_files=["/path/to/other/files/to/include/in/tarball.pickle"],\n    s3_path="s3://bucket/models/model-name/date-version/",\n)\n```\n\n## Contributing\n\nContributions are very welcome. To learn more, see the [Contributor Guide].\n\n## License\n\nDistributed under the terms of the [Apache 2.0 license][license], _Transformer Embeddings_ is free and open source software.\n\n## Issues\n\nIf you encounter any problems, please [file an issue] along with a detailed description.\n\n## Credits\n\nThis project was partly generated from [@cjolowicz]\'s [Hypermodern Python Cookiecutter] template.\n\n[@cjolowicz]: https://github.com/cjolowicz\n[pypi]: https://pypi.org/\n[hypermodern python cookiecutter]: https://github.com/cjolowicz/cookiecutter-hypermodern-python\n[file an issue]: https://github.com/ginger-io/transformer-embeddings/issues\n[pip]: https://pip.pypa.io/\n\n<!-- github-only -->\n\n[license]: https://github.com/ginger-io/transformer-embeddings/blob/main/LICENSE\n[contributor guide]: https://github.com/ginger-io/transformer-embeddings/blob/main/CONTRIBUTING.md\n[command-line reference]: https://transformer-embeddings.readthedocs.io/en/latest/usage.html\n',
    'author': 'Headspace Health',
    'author_email': 'transformer-embeddings@headspace.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/ginger-io/transformer-embeddings',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)



from torch import Tensor, clamp
from torch import sum as torch_sum
from transformers import BatchEncoding
from transformers.file_utils import ModelOutput


def mean_pooling(model_output: ModelOutput, model_inputs: BatchEncoding) -> Tensor:
    """
    Mean pooling for the model output.

    This is the unweighted average of the token embeddings, while ignoring padding
    and padded tokens.
    Copied from: https://huggingface.co/sentence-transformers/msmarco-distilroberta-base-v2
    Comments ours :).

    Parameters
    ----------
    model_output : ModelOutput
        Output from the model.
    model_inputs : BatchEncoding
        Encoded, tokenized input to the model.

    Returns
    -------
    Tensor
        Mean pooled output.
    """
    # last_hidden_state is the per token embedding.
    last_hidden_state = model_output.last_hidden_state
    attention_mask = model_inputs.attention_mask
    # Expand attention_mask (2D) to the shape of the last_hidden_state (3D).
    attention_mask_expanded = (
        attention_mask.unsqueeze(-1).expand(last_hidden_state.size()).float()
    )
    # Multiply with the expanded attention mask to make the
    # last_hidden_state for padding tokens to 0.
    sum_embeddings = torch_sum(last_hidden_state * attention_mask_expanded, dim=1)
    sum_mask = clamp(attention_mask_expanded.sum(1), min=1e-9)
    return sum_embeddings / sum_mask


def get_pooler_output(model_output: ModelOutput, model_inputs: BatchEncoding) -> Tensor:
    """
    Return the pooler output.

    Parameters
    ----------
    model_output : ModelOutput
        Output from the model.
    model_inputs : BatchEncoding
        Encoded, tokenized input to the model. Not used in this function.

    Returns
    -------
    Tensor
        Pooler output.
    """
    return model_output.pooler_output
